/*
 * tamba_model_rx_intf_enet.v
 *
 * Description:
 *	Tamba Rx Interface Ethernet Stream model
 *
 */

module tamba_model_rx_intf_enet #
  (
     /* personality
      */
    parameter SBC  = 64,
    parameter PWC  = 1,
    parameter BPW  = 8,

     /* features
      */
    parameter CFG_CHECK_IDLE  = 0,

     /* log2
      */
    parameter log2SBC  = ((SBC>256)?9:
			  (SBC>128)?8:
			  (SBC> 64)?7:
			  (SBC> 32)?6:
			  (SBC> 16)?5:
			  (SBC>  8)?4:
			  (SBC>  4)?3:
			  (SBC>  2)?2:1),

     /* int
      */
    parameter PBC     = PWC*BPW,

    parameter ST_SOP  = 0,
    parameter ST_PKT  = 1
   )
  (
    input                i_clk,
    input 		 i_sreset,

     /* Configuration
      */
    input [15:0] 	 i_cfg_size_min,
    input [15:0] 	 i_cfg_size_max,
    input [15:0] 	 i_cfg_rep,
    input 		 i_cfg_preamble_en,
    input [2:0] 	 i_cfg_preamble_wc,
    input 		 i_cfg_pad,
    input [7:0] 	 i_cfg_pad_size,
    input 		 i_cfg_crc,

     /* Control
      */
    input 		 i_start_strb,
    output 		 o_first_strb,		 
    output 		 o_end_strb,
    output 		 o_end_ok,

     /* Stream Interface
      */
    input 		 i_rx_en,
    input 		 i_rx_sop,
    input 		 i_rx_eop,
    input 		 i_rx_err,
    input [7:0] 	 i_rx_bc,
    input [PBC*8-1:0] 	 i_rx_preamble,	 
    input [SBC*8-1:0] 	 i_rx_data
   ); 

/*************************************************************************************
 * Register declarations 
 * 
 *************************************************************************************/
   reg 		       rx_st_n;
   reg 		       rx_st_r;

   reg 		       rx_err_bus_protocol_r;
   reg 		       rx_err_bus_protocol_n;
   
   wire 	       end_strb;
   reg 		       end_strb_r;

   reg 		       rx_en_r;
   reg 		       rx_sop_r;
   reg 		       rx_eop_r;
   reg 		       rx_err_r;
   reg [log2SBC-1:0]   rx_bc_r;
   reg [PBC*8-1:0]     rx_preamble_r;
   reg [SBC*8-1:0]     rx_data_r;
   
   reg 		       error_alignment_r;
   reg 		       error_idle_r;
   reg 		       error_preamble_r;
   reg 		       error_data_r;
   reg 		       error_eope_r;
   reg 		       error_r;
   reg 		       first_r;
   
   wire [7:0] 	       tx_bc;
   wire [PBC*8-1:0]    tx_preamble;
   wire [SBC*8-1:0]    tx_data;
   
   reg [7:0] 	       tx_db[SBC-1:0];
   reg [7:0] 	       rx_db_r[SBC-1:0];
   reg [SBC-1:0]       data_eq_n;

   integer 	       i;
      
/*************************************************************************************
 * Always, bus protocol state
 * 
 *************************************************************************************/
   always @(*)
     begin
	rx_st_n  = rx_st_r;
	
	if (i_rx_en)
	  begin
	     case (rx_st_r)
	       ST_SOP:
		 begin
		    case ({i_rx_sop,i_rx_eop})
		      2'b00:    rx_st_n  = ST_PKT;
		      2'b01:    rx_st_n  = ST_SOP;
		      2'b10:    rx_st_n  = ST_PKT;
		      default:  rx_st_n  = ST_SOP;
		    endcase
		 end
	       default: /*ST_PKT:*/
		 begin
		    case ({i_rx_sop,i_rx_eop})
		      2'b00:    rx_st_n  = ST_PKT;
		      2'b01:    rx_st_n  = ST_SOP;
		      2'b10:    rx_st_n  = ST_PKT;
		      default:  rx_st_n  = ST_SOP;
		    endcase
		 end
	     endcase
	  end
     end

   always @(posedge i_clk)
     if (i_sreset)
       rx_st_r  <= ST_SOP;
     else
       rx_st_r  <= rx_st_n;
	  
/*************************************************************************************
 * Always, bus protocol error
 * 
 *************************************************************************************/
   always @(*)
     begin
	rx_err_bus_protocol_n = 1'b0;
	
	if (i_rx_en)
	  begin
	     casex (rx_st_r)
	       ST_SOP:
		 begin
		    casex ({i_rx_sop,i_rx_eop})
		      2'b00:    rx_err_bus_protocol_n  = 1'b1;
		      2'b01:    rx_err_bus_protocol_n  = 1'b1;
		      2'b10:    rx_err_bus_protocol_n  = 1'b0;
		      default:  rx_err_bus_protocol_n  = 1'b0;
		    endcase
		 end
	       default: /*ST_PKT:*/
		 begin
		    casex ({i_rx_sop,i_rx_eop})
		      2'b00:    rx_err_bus_protocol_n  = 1'b0;
		      2'b01:    rx_err_bus_protocol_n  = 1'b0;
		      2'b10:    rx_err_bus_protocol_n  = 1'b1;
		      default:  rx_err_bus_protocol_n  = 1'b1;
		    endcase
		 end
	     endcase
	  end
     end

   always @(posedge i_clk)
     if (i_sreset)
       rx_err_bus_protocol_r  <= 1'b0;
     else
       rx_err_bus_protocol_r  <= rx_err_bus_protocol_n;
   
   always @(posedge i_clk)
     if (rx_err_bus_protocol_r)
       begin
	  $display("---------- ERROR -----------");
	  $display("Time: %t  ",$time);
	  $display("Module: %m");
	  $display("Description: Bus Protocol Error.\n");
       end

/*************************************************************************************
 * Always, delay 
 * 
 *************************************************************************************/
   always @(posedge i_clk)
     if (i_sreset)
       begin
	  rx_en_r        <= 1'b0;
       end
     else
       begin
	  rx_en_r        <= i_rx_en;
	  rx_sop_r       <= i_rx_sop;
	  rx_eop_r       <= i_rx_eop;
	  rx_err_r       <= i_rx_err;
	  rx_bc_r        <= i_rx_bc[0+:log2SBC];
	  rx_preamble_r  <= i_rx_preamble;
	  rx_data_r      <= i_rx_data;
       end
   
/*************************************************************************************
 * Always, first
 * 
 *************************************************************************************/
   always @(posedge i_clk)
     if (i_sreset)
       begin
	  first_r  <= 1'b1;
       end
     else
       begin
	  if (i_rx_en & i_rx_sop)
	    first_r  <= 1'b0;
	  else if (o_end_strb)
	    first_r  <= 1'b1;
       end
   
/*************************************************************************************
 * Stream Tx
 * 
 *************************************************************************************/
   tamba_model_tx_intf_enet #
     (
      .SBC      (SBC),
      .log2SBC  (log2SBC),
      .PWC      (PWC),
      .BPW      (BPW)
      )
     model_tx
       (
	.i_clk           (i_clk),
	.i_sreset        (i_sreset),
      
          /* cfg
	   */
	.i_cfg_size_min  (1'b0 & i_cfg_preamble_en ? i_rx_preamble[8+:16] : i_cfg_size_min),
	.i_cfg_size_max  (1'b0 & i_cfg_preamble_en ? i_rx_preamble[8+:16] : i_cfg_size_max),
	.i_cfg_rep       (1'b0 & i_cfg_preamble_en ? 16'h1                : i_cfg_rep),
	.i_cfg_ipg       (8'h0),
	.i_cfg_pad       (i_cfg_pad),
	.i_cfg_pad_size  (i_cfg_pad_size),
	.i_cfg_crc       (i_cfg_crc),
	
          /* Control
	   */
	.i_start_strb    (1'b0 & i_cfg_preamble_en ? (i_rx_en & i_rx_sop) : i_start_strb),
	.o_first_strb    (o_first_strb),
	.o_end_strb      (end_strb),
	
          /* Stream Interface
	   */
	.i_fc            (!i_rx_en),
	
	.o_en            (tx_en),
	.o_sop           (tx_sop),
	.o_eop           (tx_eop),
	.o_eope          (tx_eope),
	.o_bc            (tx_bc),
	.o_preamble      (tx_preamble),
	.o_data          (tx_data)
	); 

   always @(posedge i_clk)
     end_strb_r  <= end_strb;

/*************************************************************************************
 * Check alignment
 * 
 *************************************************************************************/
   always @(posedge i_clk)
     if (i_sreset)
       error_alignment_r  <= 1'b0;
     else
       begin
	  if (rx_en_r & rx_sop_r & !error_alignment_r)
	    begin
	       if (rx_data_r[(SBC-1)*8+:8] !== 8'h1)
		 begin
		    error_alignment_r  <= 1'b1;
		    
		    $display("---------- ERROR -----------");
		    $display("Time: %t  ",$time);
		    $display("Module: %m");
		    $display("Description: Bad Data Alignment.\n");
		 end
	    end
       end

/*************************************************************************************
 * Check idle
 * 
 *************************************************************************************/
   always @(posedge i_clk)
     if (i_sreset)
       error_idle_r  <= 1'b0;
     else if (CFG_CHECK_IDLE==1)
       begin
	  casex({i_rx_en,i_rx_sop,i_rx_eop,i_rx_err})
	    4'b1x_xx, 4'b00_00: ;
	    default:
	      begin
		 error_idle_r  <= 1'b1;
		 
		 $display("---------- ERROR -----------");
		 $display("Time: %t  ",$time);
		 $display("Module: %m");
		 $display("Description: Invalid Control Signals during idle.\n");
	      end
	  endcase
       end
   
/*************************************************************************************
 * Check Preamble
 * 
 *************************************************************************************/
   always @(posedge i_clk)
     if (i_sreset)
       error_preamble_r  <= 1'b0;
     else
       begin
	  if (rx_en_r & rx_sop_r & !error_preamble_r & i_cfg_preamble_en)
	    begin
	       if (tx_preamble[0+:((PBC-1)*8)] != rx_preamble_r[0+:((PBC-1)*8)])
		 begin
		    error_preamble_r  <= 1'b1;
		    
		    $display("---------- ERROR -----------");
		    $display("Time: %t  ",$time);
		    $display("Module: %m");
		    $display("Description: Bad Preamble.\n");
		 end
	    end
       end

/*************************************************************************************
 * Check Data
 * 
 *************************************************************************************/
   always @(*)
     for (i=0;i<SBC;i=i+1)
       tx_db[i][7:0]  = tx_data[(i*8) +: 8];

   always @(*)
     for (i=0;i<SBC;i=i+1)
       rx_db_r[i][7:0]  = rx_data_r[(i*8) +: 8];

   always @(*)
     begin
	if (tx_en)
	  for (i=0;i<SBC;i=i+1)
	    data_eq_n[i]  = ((tx_db[i] === rx_db_r[i]) | (i < ((SBC-1) - tx_bc[0+:log2SBC])));
     end

   always @(posedge i_clk)
     if (i_sreset)
       error_data_r  <= 1'b0;
     else
       begin
	  if (rx_en_r & !error_data_r)
	    begin
	       if (((tx_en              !== rx_en_r)   |
		    (tx_sop             !== rx_sop_r)  |
		    (tx_eop             !== rx_eop_r)  |
		    (tx_bc[0+:log2SBC]  !== rx_bc_r)   |
		    (!(&data_eq_n)))     & !rx_err_r)
		 begin
		    error_data_r  <= 1'b1;
		    
		    $display("---------- ERROR -----------");
		    $display("Time: %t  ",$time);
		    $display("Module: %m");
		    $display("Description: Bad Data.\n");
		 end
	    end
       end

/*************************************************************************************
 * Check eope
 * 
 *************************************************************************************/
   always @(posedge i_clk)
     if (i_sreset)
       begin
	  error_eope_r  <= 1'b0;
       end
     else
       begin
	  if (rx_en_r & (rx_err_r !== 1'b0) & !tx_eope & !error_eope_r)
	    begin
	       error_eope_r  <= 1'b1;
	       
	       $display("---------- ERROR -----------");
	       $display("Time: %t  ",$time);
	       $display("Module: %m");
	       $display("Description: EOPE Error Event.\n");
	    end
       end
   
/*************************************************************************************
 * always. error_
 * 
 *************************************************************************************/
   always @(*)
     error_r  = (error_alignment_r | error_preamble_r | error_data_r | error_eope_r);
   
/*************************************************************************************
 * Assign
 * 
 *************************************************************************************/
   assign o_end_strb  = end_strb_r;
   assign o_end_ok    = !error_r;

endmodule

